/*
 * usbcore.c:
 *   by uratan! 2008.9.18
 */
#include "common.h"
#include "usb.h"
#include "usbhw.h"

/*
 * my local debug flag
 */
#define UDEBUG 1
#define rPRINTF  if(UDEBUG)rprintf

/*
 * internal buffer size for endpoint 0
 *   must be greater than 'max_packet_size_ep0' described in device descriptor
 */
#define EP0_BUFF_SIZE    64

/*
 * Max Number of Interfaces
 */
#define MAX_N_IF          4

/*
 * make state machine for CONTROL-PIPE
 */
#define CPS_IDLE     0		/* only SETUP will be accepted */
#define CPS_DATAIN   1		/* expect IN from host */
#define CPS_DATAOUT  2		/* expect OUT from host */
#define CPS_STATOUT  3		/* expect OUT from host */
#define CPS_STATIN   4		/* expect IN from host */
#if 0
[SUDS_IDLE]
 | SETUP(OUT)
 |     WITH DATA
				[SUDS_DATAOUT]
 |         IN->stall
 |         OUT
 |             ->stall if buffer over
 |             ->recv data until specified length--> ep0_data_ready=TRUE
						[SUDS_STATIN]
 |                 IN->send status
 |                 OUT->stall
 |     WITHOUT DATA
				[SUDS_STATIN] <========
 |         IN->send status
 |         OUT->stall
 | SETUP(IN)
 |     WITHOUT DATA
				[ucore_STALL_EP0()]
 |         IN->stall
 |         OUT->stall
 |     WITH DATA
				[SUDS_DATAIN] <=========
 |         OUT->stall
 |         IN->send data (zero-length-packet if needed)
										(datasize<requestsize) &&
										(datasize % packetsize == 0)
					[SUDS_STATOUT]
 |             OUT->send status
 |             IN->stall
#endif /* 0 */

/*
 *
 */
#define EP_CTRL_IN   (0x80 | 0)
#define EP_CTRL_OUT  (0x00 | 0)

/*
 * functions
 */
int ucore_InitialCore(void);
void ucore_ResetCore(void);
void ucore_WakeUp_host(void);
void ucore_SetupCommand(void);
void ucore_start_DataIn(u_char *buff, u_long len);
void ucore_start_DataOut(u_char *buff, u_long len, u_long bufsize);
void ucore_wait_DataOut_finished(void);
void ucore_start_StatusIn(void);
static int ucore_GetStatus(void);
static int ucore_SetClrFeature(u_long sc);
static int ucore_GetDescriptor(void);
static u_char *ucore_Get_NthConfigDscr(u_char index);
static u_char *ucore_Get_NthStringDscr(u_char index);
static int ucore_SetConfiguration(void);
static int ucore_SetInterface(void);
void ucore_isr_Ep0(u_long event);
static void ucore_DataInStage(void);
static int ucore_DataOutStage(void);
static void ucore_StatusInStage(void);
static void ucore_StatusOutStage(void);
void ucore_STALL_EP0(void);

/*
 * variables
 */
u_char *USB_DeviceDescriptor;
u_char *USB_ConfigDescriptor;
u_char *USB_StringDescriptor;
u_long max_packet_size_ep0;
u_short def_device_status;
u_short USB_DeviceStatus;
u_char  USB_Configuration;	/* 0 or configuration number */
u_char  USB_NumInterfaces;	/* of selected configuration */
u_char  USB_AltSetting[MAX_N_IF];

/*static*/ u_char  ucore_device_addr;
/*static*/ u_long  ucore_bmEP_enabled;
/*static*/ u_long  ucore_bmEP_stalled;

/*static*/ int control_pipe_state;
/*static*/ int cpipe_add_zero_packet;
/*static*/ int volatile cpipe_out_data_ready;
/*static*/ u_long cpipe_len_requested;
/*static*/ u_char *cpipe_data;
/*static*/ u_long cpipe_count;
/*static*/ u_long cpipe_out_buff_remain;	/* OUT only */

u_char SETUPDAT[8];
int GotSUD;
u_char EP0BUF[EP0_BUFF_SIZE];

/*
 * externs
 */
/* in dscr.S */
extern u_long *DeviceDscr, *ConfigDscr, *StringDscr;
/* in usbhw.c */
extern int uhw_Init(void);
extern void uhw_WakeUp(void);
extern void uhw_SetAddress(u_char adr);
extern void uhw_Configure(int cfg);
extern void uhw_ConfigEP(u_char *pEPD);
extern void uhw_EnableEP(u_char EPNum);
extern void uhw_DisableEP(u_char EPNum);
extern void uhw_ResetEP(u_char EPNum);
extern void uhw_SetStallEP(u_char EPNum);
extern void uhw_ClrStallEP(u_char EPNum);
extern u_long uhw_WriteEP(u_char EPNum, u_char *pData, u_long cnt);
extern void uhw_arm_WriteEP(u_char EPNum);
extern int uhw_ReadEP(u_char EPNum, u_char *pData, u_long DataSize);
extern void uhw_ReadEP_finish(u_char EPNum);
/* in vcomuser.c */
extern void vcom_USB_Configure_Event(void);
extern void vcom_USB_Interface_Event(void);
extern void vcom_USB_Feature_Event(void);
/* in periph.c */
extern int DR_SetupCommand(void);
extern int DR_GetDescriptor(void);


/*
 * initial valiables about USB core
 */
int ucore_InitialCore(void)
{
	USB_DeviceDescriptor = (u_char *)&DeviceDscr;
	USB_ConfigDescriptor = (u_char *)&ConfigDscr;
	USB_StringDescriptor = (u_char *)&StringDscr;

	max_packet_size_ep0 = USB_DeviceDescriptor[7];

	if(USB_ConfigDescriptor[/*bmAttributes*/ 7] & UC_SELF_POWERED)
		def_device_status = UDS_SELF_POWERED;
	else
		def_device_status = 0x00;

	control_pipe_state = CPS_IDLE;

	/*
	 * initial usbhw.c (max_packet_size_ep0 needed)
	 */
	return(uhw_Init());
}

/*
 * Reset USB Core
 */
void ucore_ResetCore(void)
{
	USB_DeviceStatus = def_device_status;
/* has UDS_SELF_POWERED, UDS_REMOTE_WAKEUP */
	USB_Configuration = 0;
/* has configuration number */
	ucore_device_addr = 0;

#define bmEP_IN(x)  ((1<<16)<<(x))
#define bmEP_OUT(x)  (1<<(x))
#define bmEP_EPnum(n)  (((n)&0x80) ? bmEP_IN((n)&0x0f) : bmEP_OUT((n)&0x0f))
#define DEF_bmEP_ENABLED  (bmEP_IN(0) | bmEP_OUT(0))
	ucore_bmEP_enabled = DEF_bmEP_ENABLED;		/*  */
	ucore_bmEP_stalled = 0x00000000;			/*  */
}

/*
 * USB Remote Wakeup Function
 */
void ucore_WakeUp_host(void)
{
	if(USB_DeviceStatus & UDS_REMOTE_WAKEUP) {
		uhw_WakeUp();
	}
}

/*
 * device request parser
 */
void ucore_SetupCommand(void)
{
	/*
	 * take SETUP data out from hardware buffer
	 */
	uhw_ReadEP(EP_CTRL_OUT, SETUPDAT, sizeof(SETUPDAT));

	/*
	 * clear the flag (before release the hardware buffer)
	 */
	GotSUD = FALSE;
//	uhw_ReadEP_finish(EP_CTRL_OUT);
		/*
		 * I will release the buffer if needed in the following process.
		 * the host receives NAK by next OUT packet and wait until
		 * we are ready(==release-buffer).
		 */
	/*
	 * sample some params in the future
	 */
	cpipe_len_requested = SETUPDAT[6] + (SETUPDAT[7] << 8);

#if UDEBUG
	{
		int k;

		rprintf("SETUP[]:");
		for(k=0; k<8; k++) {
			rprintf(" 0x%02x", SETUPDAT[k]);
		}
		rprintf("\r\n");
	}
#endif /* UDEBUG */

	/*
	 * parse the request
	 * at first, parse device-dependant requests here
	 */
  if(DR_SetupCommand() != DONE) {
	switch(SETUPDAT[0] & UT_TYPE_MASK) {
	case UT_STANDARD:
			switch(SETUPDAT[1]) {
			case UR_GET_STATUS:
					rPRINTF("GET_STAT\r\n");
					if(ucore_GetStatus() != DONE) {
						goto stall_i;
					}
					break;

			case UR_CLEAR_FEATURE:
					rPRINTF("CLR_FEAT\r\n");
					if(ucore_SetClrFeature(0) != DONE) {
						goto stall_i;
					}
//					vcom_USB_Feature_Event();
					break;

			case UR_SET_FEATURE:
					rPRINTF("SET_FEAT\r\n");
					if(ucore_SetClrFeature(1) != DONE) {
						goto stall_i;
					}
//					vcom_USB_Feature_Event();
					break;

			case UR_SET_ADDRESS:
					rPRINTF("SET_ADDR\r\n");
					switch(SETUPDAT[0] & UT_RECIP_MASK) {
					case UT_TO_DEVICE:
							ucore_device_addr = 0x80 | SETUPDAT[2];
							ucore_start_StatusIn();
							break;
					default:
							goto stall_i;
					}
					break;

			case UR_GET_DESCRIPTOR:
				rPRINTF("GET_DESC\r\n");
				if(DR_GetDescriptor() != DONE)
					if(ucore_GetDescriptor() != DONE) {
						goto stall_i;
					}
					break;

			case UR_SET_DESCRIPTOR:
					rPRINTF("SET_DESC\r\n");
					ucore_STALL_EP0();
//					cpipe_count = 0;
					break;

			case UR_GET_CONFIGURATION:
					rPRINTF("GET_CONF\r\n");
					switch(SETUPDAT[0] & UT_RECIP_MASK) {
					case UT_TO_DEVICE:
							EP0BUF[0] = USB_Configuration;
							ucore_start_DataIn(EP0BUF, 1);
							break;
					default:
							goto stall_i;
					}
					break;

			case UR_SET_CONFIGURATION:
					rPRINTF("SET_CONF\r\n");
					switch(SETUPDAT[0] & UT_RECIP_MASK) {
					case UT_TO_DEVICE:
							if(ucore_SetConfiguration() != DONE) {
								goto stall_i;
							}
//							vcom_USB_Configure_Event();
							break;
					default:
							goto stall_i;
					}
					break;

			case UR_GET_INTERFACE:
					rPRINTF("GET_INTF\r\n");
					switch(SETUPDAT[0] & UT_RECIP_MASK) {
					case UT_TO_INTERFACE:
							if(
								(USB_Configuration != 0) &&
								(SETUPDAT[4] < USB_NumInterfaces)
							) {
								EP0BUF[0] = USB_AltSetting[SETUPDAT[4]];
								ucore_start_DataIn(EP0BUF, 1);
							} else {
								goto stall_i;
							}
							break;
					default:
							goto stall_i;
					}
					break;

			case UR_SET_INTERFACE:
					rPRINTF("SET_INTF\r\n");
					switch(SETUPDAT[0] & UT_RECIP_MASK) {
					case UT_TO_INTERFACE:
							if(ucore_SetInterface() != DONE) {
								goto stall_i;
							}
//							vcom_USB_Interface_Event();
							break;
					default:
							goto stall_i;
					}
					break;

			default:
					goto stall_i;
			}
			break;

	case UT_CLASS:
	case UT_VENDOR:
	default:
stall_i:
			ucore_STALL_EP0();
//			cpipe_count = 0;
			break;
	}
  }
	rPRINTF("@@@\r\n");
}

/*
 * initiate data IN
 */
void ucore_start_DataIn(u_char *buff, u_long len)
{
	/*
	 * setup control pipe params
	 */
	cpipe_data = buff;
	cpipe_count = len;

	/*
	 * round data length by request length
	 */
	if(cpipe_count > cpipe_len_requested) {
		cpipe_count = cpipe_len_requested;
	}

	/*
	 * I expect IN from the host
	 */
	control_pipe_state = CPS_DATAIN;
	cpipe_add_zero_packet = FALSE;		/* for IN data */

	/*
	 * parse the request finished, but keep OUT buffer filled,
	 * no OUT packet should come until status-out phase
	 */
//	uhw_ReadEP_finish(EP_CTRL_OUT);

	/*
	 * initiate first data-in packet
	 */
	ucore_DataInStage();
}

/*
 * initiate data OUT
 */
void ucore_start_DataOut(u_char *buff, u_long len, u_long bufsize)
{
	/*
	 * setup control pipe params
	 */
	cpipe_data = buff;
	cpipe_out_buff_remain = bufsize;
	cpipe_count = len;

	/*
	 * round data length by request length
	 */
	if(cpipe_count > cpipe_len_requested) {
		cpipe_count = cpipe_len_requested;
	}

	/*
	 * I expect OUT from the host
	 */
	control_pipe_state = CPS_DATAOUT;
	cpipe_out_data_ready = FALSE;		/* for OUT data */

	/*
	 * parse the request finished, make out buffer free
	 */
	uhw_ReadEP_finish(EP_CTRL_OUT);

	/*
	 * wait for OUT packet will come
	 */
}

/*
 * wait for OUT data coming
 */
void ucore_wait_DataOut_finished(void)
{
	/*
	 * wait for OUT packet will come
	 */
	while(1) {
		if(cpipe_out_data_ready)
			break;
	}
		/* timeout ? */
		/* how handle USB-RESET or etc ? */
}

/*
 * initiate status IN
 */
void ucore_start_StatusIn(void)
{
	/*
	 * I expect IN from the host
	 */
	control_pipe_state = CPS_STATIN;

	/*
	 * parse the request finished, but keep OUT buffer filled,
	 * no OUT packet should come until status-out phase
	 */
//	uhw_ReadEP_finish(EP_CTRL_OUT);

	/*
	 * initiate status-in packet
	 */
	ucore_StatusInStage();
}

/*
 * Get Status USB Request
 */
static int ucore_GetStatus(void)
{
	u_long n, m;

	switch(SETUPDAT[0] & UT_RECIP_MASK) {
	case UT_TO_DEVICE:
			EP0BUF[0] = (USB_DeviceStatus>>0) & 0xff;
			EP0BUF[1] = (USB_DeviceStatus>>8) & 0xff;
			break;

	case UT_TO_INTERFACE:
			if((USB_Configuration != 0) && (SETUPDAT[4] < USB_NumInterfaces)) {
				EP0BUF[0] = 0;
				EP0BUF[1] = 0;
			} else {
				return(YET);
			}
			break;

	case UT_TO_ENDPOINT:
			n = SETUPDAT[4] & 0x8F;
			m = bmEP_EPnum(n);
			if(
				((USB_Configuration != 0) || ((n & 0x0F) == 0)) &&
				(ucore_bmEP_enabled & m)
			) {
				EP0BUF[0] = (ucore_bmEP_stalled & m) ? 1 : 0;
/*/* shall sync with stall operation ! */
				EP0BUF[1] = 0;
			} else {
				return(YET);
			}
			break;

	default:
			return(YET);
	}

	ucore_start_DataIn(EP0BUF, 2);
	return(DONE);
}

/*
 * Set/Clear Feature USB Request
 *   pass sc as: 0 for Clear, 1 for Set
 */
static int ucore_SetClrFeature(u_long sc)
{
	u_long n, m;

	switch(SETUPDAT[0] & UT_RECIP_MASK) {
	case UT_TO_DEVICE:
			if((SETUPDAT[2]+(SETUPDAT[3]<<8)) == UF_REMOTE_WAKEUP) {
				if(sc) {
					USB_DeviceStatus |=  UDS_REMOTE_WAKEUP;
				} else {
					USB_DeviceStatus &= ~UDS_REMOTE_WAKEUP;
				}
			} else {
				return(YET);
			}
			break;

	case UT_TO_INTERFACE:
			return(YET);

	case UT_TO_ENDPOINT:
			n = SETUPDAT[4] & 0x8F;
			m = bmEP_EPnum(n);
			if(
				(USB_Configuration != 0) &&
				((n & 0x0F) != 0) &&
				(ucore_bmEP_enabled & m)
			) {
				if((SETUPDAT[2]+(SETUPDAT[3]<<8)) == UF_ENDPOINT_HALT) {
					if(sc) {
						uhw_SetStallEP(n);
						ucore_bmEP_stalled |=  m;
					} else {
						uhw_ClrStallEP(n);
						ucore_bmEP_stalled &= ~m;
					}
				} else {
					return(YET);
				}
			} else {
				return(YET);
			}
			break;

	default:
			return(YET);
	}

	ucore_start_StatusIn();
	return(DONE);
}

/*
 * Get Descriptor USB Request
 */
static int ucore_GetDescriptor(void)
{
	u_char *dscr;
	u_long len;

	switch(SETUPDAT[0] & UT_RECIP_MASK) {
	case UT_TO_DEVICE:
			switch(SETUPDAT[3]) {
			case UDESC_DEVICE:
					rPRINTF("UDESC_DEVICE\r\n");
					dscr = USB_DeviceDescriptor;
					len = dscr[/*bLength*/ 0];
					break;

			case UDESC_CONFIG:
					rPRINTF("UDESC_CONFIG\r\n");
					dscr = ucore_Get_NthConfigDscr(SETUPDAT[2]);
					if(dscr == NULL) {
						return(YET);
					}
					len = dscr[2] + (dscr[3]<<8);	/* wTotalLength */
					break;

			case UDESC_STRING:
					rPRINTF("UDESC_STRING\r\n");
					dscr = ucore_Get_NthStringDscr(SETUPDAT[2]);
					if(dscr == NULL) {
						return(YET);
					}
					len = dscr[/*bLength*/ 0];		/* bLength */
					break;

			default:
					return(YET);
			}
			break;

	case UT_TO_INTERFACE:
	default:
			return(YET);
	}

	ucore_start_DataIn(dscr, len);
	return(DONE);
}

/*
 * search Nth config descriptor
 */
static u_char *ucore_Get_NthConfigDscr(u_char index)
{
	u_char *dscr;

	dscr = USB_ConfigDescriptor;

	while(dscr[1 /*type*/] == UDESC_CONFIG) {
		if(index == 0)
			return(dscr);
		index--;
		dscr += dscr[2] + (dscr[3]<<8);		/* wTotalLength */
	}

	return(NULL);
}

/*
 * search Nth string descriptor
 */
static u_char *ucore_Get_NthStringDscr(u_char index)
{
	u_char *dscr;

	dscr = USB_StringDescriptor;

	while(dscr[1 /*type*/] == UDESC_STRING) {
		if(index == 0)
			return(dscr);
		index--;
		dscr += dscr[0 /*length*/];
	}

	return(NULL);
}

/*
 * Set Configuration USB Request
 */
static int ucore_SetConfiguration(void)
{
	u_char *dscr;
	u_long alt, n, m;

	if(SETUPDAT[2]) {
		/*
		 * loop for all configuration descriptors
		 *   don't forget to add '0x00' at
		 *   the end of configuration descriptors !
		 */
		dscr = USB_ConfigDescriptor;

		while(dscr[/*bLength*/ 0]) {
			switch(dscr[/*bDescriptorType*/ 1]) {
			case UDESC_CONFIG:
					/*
					 * search requested configuration descriptor set
					 */
					if(dscr[/*bConfigurationValue*/ 5] == SETUPDAT[2]) {
						/*
						 * match
						 */
						USB_Configuration = SETUPDAT[2];
						USB_NumInterfaces = dscr[/*bNumInterfaces*/ 4];

						/*
						 * initial misc params/hardwares
						 */
						for(n=0; n<MAX_N_IF; n++) {
							USB_AltSetting[n] = 0;
						}
						for(n=1; n<16; n++) {
							if(ucore_bmEP_enabled & bmEP_OUT(n)) {
								uhw_DisableEP(n);
							}
							if(ucore_bmEP_enabled & bmEP_IN(n)) {
								uhw_DisableEP(n | 0x80);
							}
						}
						ucore_bmEP_enabled = DEF_bmEP_ENABLED;
						ucore_bmEP_stalled = 0x00000000;
						uhw_Configure(TRUE);

						if(dscr[/*bmAttributes*/ 7] & UC_SELF_POWERED) {
							USB_DeviceStatus |=  UDS_SELF_POWERED;
						} else {
							USB_DeviceStatus &= ~UDS_SELF_POWERED;
						}
					} else {
						/*
						 * point next configuration descriptor
						 */
						dscr += dscr[2] + (dscr[3]<<8);	/* wTotalLength */
						continue;
					}
					break;

			case UDESC_INTERFACE:
					alt = dscr[/*bAlternateSetting*/ 3];
					break;

			case UDESC_ENDPOINT:
					if(alt == 0) {
						/*
						 * enable endpoints about default setting
						 */
						n = dscr[/*bEndpointAddress*/ 2] & 0x8F;
						m = bmEP_EPnum(n);
						ucore_bmEP_enabled |= m;
						uhw_ConfigEP(dscr);
						uhw_EnableEP(n);
						uhw_ResetEP(n);
					}
					break;
			}

			dscr += dscr[/*bLength*/ 0];
		}
	} else {
		/*
		 * configuration canceled
		 */
		USB_Configuration = 0;

		for(n=1; n<16; n++) {
			if(ucore_bmEP_enabled & bmEP_OUT(n)) {
				uhw_DisableEP(n);
			}
			if(ucore_bmEP_enabled & bmEP_IN(n)) {
				uhw_DisableEP(n | 0x80);
			}
		}
		ucore_bmEP_enabled  = DEF_bmEP_ENABLED;
		ucore_bmEP_stalled  = 0x00000000;
		uhw_Configure(FALSE);
	}

	if(USB_Configuration == SETUPDAT[2]) {
		ucore_start_StatusIn();
		return(DONE);
	} else {
		return(YET);
	}
}

/*
 * Set Interface USB Request
 */
static int ucore_SetInterface(void)
{
	int set;
	u_char *dscr;
	u_char ifn, alt, old;
	u_long msk, n, m;

	/*
	 * am I configured ?
	 */
	if(USB_Configuration == 0)
		return(YET);

	set = FALSE;

	/*
	 * loop for all configuration descriptors
	 *   don't forget to add '0x00' at
	 *   the end of configuration descriptors !
	 */
	dscr = USB_ConfigDescriptor;

	while(dscr[/*bLength*/ 0]) {
		switch(dscr[/*bDescriptorType*/ 1]) {
		case UDESC_CONFIG:
				/*
				 * search requested configuration descriptor set
				 */
				if(dscr[/*bConfigurationValue*/ 5] != USB_Configuration) {
					/*
					 * point next configuration descriptor
					 */
					dscr += dscr[2] + (dscr[3]<<8);		/* wTotalLength */
					continue;
				}
				break;

		case UDESC_INTERFACE:
				/*
				 * is this descriptor match for requested ?
				 */
				ifn = dscr[/*bInterfaceNumber*/ 2];
				alt = dscr[/*bAlternateSetting*/ 3];
				msk = 0;
				if((ifn == SETUPDAT[4]) && (alt == SETUPDAT[2])) {
					set = TRUE;
					old = USB_AltSetting[ifn];
					USB_AltSetting[ifn] = alt;
				}
				break;

		case UDESC_ENDPOINT:
				/*
				 * it this descriptor is match for requested ?
				 */
				if(ifn == SETUPDAT[4]) {
					n = dscr[/*bEndpointAddress*/ 2] & 0x8F;
					m = bmEP_EPnum(n);
					if(alt == SETUPDAT[2]) {
						/*
						 * enable endpoints about this setting
						 */
						ucore_bmEP_enabled |=  m;
						ucore_bmEP_stalled &= ~m;
						uhw_ConfigEP(dscr);
						uhw_EnableEP(n);
						uhw_ResetEP(n);
						msk |= m;
					} else if((alt == old) && ((msk & m) == 0)) {
						/*
						 * disble endpoints about old setting...
						 */
						ucore_bmEP_enabled &= ~m;
						ucore_bmEP_stalled &= ~m;
						uhw_DisableEP(n);
							/*
							 * we scans descriptors by 1 pass.
							 * is this loop logic correct ???
							 */
					}
				}
				break;
//		default:
//				set = FALSE;
//				break;
		}

		/*
		 * point next descriptor in the configuration descriptor set
		 */
		dscr += dscr[/*bLength*/ 0];
	}

	if(set != TRUE)
		return(YET);

	ucore_start_StatusIn();
	return(DONE);
}

/* ==================================================================== */

/*
 * USB endpoint 0 interrupt handler
 */
void ucore_isr_Ep0(u_long event)
{
	switch(event) {
	case USB_EVT_SETUP:
			rPRINTF("EVT_SETUP\r\n");
			GotSUD = TRUE;
			break;

	case USB_EVT_OUT:
			rPRINTF("EVT_OUT\r\n");
			if(control_pipe_state == CPS_DATAOUT) {
				if(cpipe_count != 0) {
					if(ucore_DataOutStage() == 0) {
						if(cpipe_count == 0) {
							/*
							 * make status-in ready
							 */
//							control_pipe_state = CPS_STATIN;
//							ucore_StatusInStage();
							control_pipe_state = CPS_IDLE;
							cpipe_out_data_ready = TRUE;
						} else {
							/*
							 * permit to receive next OUT packet
							 */
							uhw_ReadEP_finish(EP_CTRL_OUT);
						}
					} else {
						/* ignore this packet */
						uhw_ReadEP_finish(EP_CTRL_OUT);
						control_pipe_state = CPS_IDLE;
						ucore_STALL_EP0();
					}
				} else {
					/* ignore this packet */
					uhw_ReadEP_finish(EP_CTRL_OUT);
					control_pipe_state = CPS_IDLE;
					ucore_STALL_EP0();
				}
			}
			else if(control_pipe_state == CPS_STATOUT) {
				/*
				 * always ignore 'status-OUT-has-come'
				 *   (following next SETUP shall be always
				 *    'Packet over-written', don't mind)
				 */
				//ucore_StatusOutStage();
				//control_pipe_state = CPS_IDLE;
			}
			else {
				control_pipe_state = CPS_IDLE;
				ucore_STALL_EP0();
			}
			break;

	case USB_EVT_IN:
			rPRINTF("EVT_IN\r\n");
			if(control_pipe_state == CPS_DATAIN) {
				if((cpipe_count != 0) || cpipe_add_zero_packet) {
					ucore_DataInStage();
				} else {
					control_pipe_state = CPS_STATOUT;
					/*
					 * permit to receive OUT packet
					 */
					uhw_ReadEP_finish(EP_CTRL_OUT);
				}
			}
			else if(control_pipe_state == CPS_STATIN) {
				/*
				 * 'status-in-finished' here, nothing to do.
				 * only about SET_ADDR request, update device address here.
				 */
				if(ucore_device_addr & 0x80) {
					ucore_device_addr &= 0x7F;
					uhw_SetAddress(ucore_device_addr);
				}
				control_pipe_state = CPS_IDLE;
			}
			else {
				control_pipe_state = CPS_IDLE;
				ucore_STALL_EP0();
			}
			break;
	}
}

/*
 * USB endpoint 0 interrupt handler for data IN 
 */
static void ucore_DataInStage(void)
{
	u_long cnt;

	/*
	 * calc data length for this packet
	 */
	if(cpipe_count > max_packet_size_ep0) {
		cnt = max_packet_size_ep0;
	} else {
		cnt = cpipe_count;
	}
	rPRINTF("DataIn:[%d] %02x %02x..\r\n", cnt, cpipe_data[0], cpipe_data[1]);

	/*
	 * write the data into the buffer
	 */
	uhw_WriteEP(EP_CTRL_IN, cpipe_data, cnt);

	/*
	 *
	 */
	if(cpipe_add_zero_packet) {
		/*
		 * already do this operation, reset this flag
		 */
		cpipe_add_zero_packet = FALSE;
	} else {
		/*
		 * advance pointers for next operation
		 */
		cpipe_data += cnt;
		cpipe_count -= cnt;
		cpipe_len_requested -= cnt;

		/*
		 * add zero-length-packet if the data length is just fit
		 * to packet size while the host expect more.
		 */
		if(
			(cnt == max_packet_size_ep0) &&
			(cpipe_count == 0) &&
			(cpipe_len_requested != 0)
		) {
			cpipe_add_zero_packet = TRUE;
		}
	}

	/*
	 * at last, arm the endpoint
	 *   (to prevent 'transfer-finished' interrupt among above process)
	 */
	uhw_arm_WriteEP(EP_CTRL_IN);
}

/*
 * USB endpoint 0 interrupt handler for data OUT
 */
static int ucore_DataOutStage(void)
{
	int cnt;

	/*
	 * take received data out
	 */
	cnt = uhw_ReadEP(EP_CTRL_OUT, cpipe_data, cpipe_out_buff_remain);
	rPRINTF("DataOut:[%d] %02x %02x..\r\n", cnt, cpipe_data[0], cpipe_data[1]);

	if(cnt < 0) {
		return(-1);		/* buffer overflow */
	}

	/*
	 * advance pointers for next operation
	 */
	cpipe_data += cnt;
	cpipe_count -= cnt;
//	cpipe_len_requested -= cnt;
	cpipe_out_buff_remain -= cnt;

	/*
	 * at last, free the endpoint
	 *   (to prevent next 'data-received' interrupt among above process)
	 */
//	uhw_ReadEP_finish(EP_CTRL_OUT);

	return(0);
}

/*
 * USB endpoint 0 interrupt handler for status IN
 */
static void ucore_StatusInStage(void)
{
	rPRINTF("StatusIn\r\n");
	uhw_WriteEP(EP_CTRL_IN, NULL, 0);
	uhw_arm_WriteEP(EP_CTRL_IN);
}

/*
 * USB endpoint 0 interrupt handler for status OUT
 */
static void ucore_StatusOutStage(void)
{
	rPRINTF("StatusOut\r\n");
	uhw_ReadEP(EP_CTRL_OUT, EP0BUF, 0);
	uhw_ReadEP_finish(EP_CTRL_OUT);
}

/*
 * stall both End Point 0 in/out
 */
void ucore_STALL_EP0(void)
{
	uhw_SetStallEP(EP_CTRL_IN);		/* for IN */
	uhw_SetStallEP(EP_CTRL_OUT);	/* for OUT */
	rPRINTF("STALL_EP0()\r\n");
		/*
		 * the stall bits are automatically cleared when next SETUP received
		 */
}
